//
// Created by Attila Haluska on 8/13/18.
//

#include "compiler/parser/parser.h"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include "compiler/parser/ast/ast_node.h"
#include "compiler/parser/precedence.h"
#include "compiler/lexer/lexer.h"
#include "compiler/lexer/token.h"
#include "memory_management/memory.h"

typedef struct parser_t {
    Token* previous_token;
    Token* current_token;
    bool had_error;
    bool panic_mode;
} Parser;

typedef struct parsing_context_t {
    Lexer lexer;
    Parser parser;
} ParsingContext;

typedef AstNode* (* NudParseFunction)(ParsingContext* context);

typedef AstNode* (* LedParseFunction)(ParsingContext* context, AstNode* left);

typedef struct parse_rule_t {
    NudParseFunction nud;
    LedParseFunction led;
    Precedence precedence;
} ParseRule;

static ParseRule* get_parse_rule(Token* token);

static Precedence get_precedence(Token* token);

static AstNode* parse_expression(ParsingContext* context);

static void parser_init(Parser* parser) {
    parser->previous_token = NULL;
    parser->current_token = NULL;
    parser->had_error = false;
    parser->panic_mode = false;
}

static void parser_free(Parser* parser) {
    DELETE_TYPE(Token, parser->current_token);
}

inline static Token* get_current_token(ParsingContext* context) {
    return context->parser.current_token;
}

inline static Token* get_previous_token(ParsingContext* context) {
    return context->parser.previous_token;
}

static void error_at_token(ParsingContext* context, const Token* token, const char* message) {
    if (context->parser.panic_mode) {
        return;
    }

    context->parser.panic_mode = true;
    context->parser.had_error = true;

    fprintf(stderr, "[line %llu] Error", token->line);

    if (token->type == TOKEN_EOF) {
        fprintf(stderr, " at the end of file");
    } else {
        fprintf(stderr, " at '%.*s'", (int) token->length, token->start);
    }

    fprintf(stderr, ": %s\n", message);
}

static void error_at_previous_token(ParsingContext* context, const char* message) {
    error_at_token(context, get_previous_token(context), message);
}

static void error_at_current_token(ParsingContext* context, const char* message) {
    error_at_token(context, get_current_token(context), message);
}

static void advance_to_next_token(ParsingContext* context) {
    context->parser.previous_token = context->parser.current_token;

    for (;;) {
        context->parser.current_token = lexer_next_token(&context->lexer);

        if (get_current_token(context)->type != TOKEN_ERROR) {
            break;
        }

        error_at_current_token(context, get_current_token(context)->error_message);
    }
}

static bool check_current_token_type(ParsingContext* context, TokenType type) {
    return get_current_token(context)->type == type;
}

static void consume_current_token_if_type(ParsingContext* context, TokenType type, const char* message) {
    if (check_current_token_type(context, type)) {
        advance_to_next_token(context);
        token_free(get_previous_token(context));
        return;
    }

    error_at_current_token(context, message);
}

static AstNode* parse_precedence(ParsingContext* context, Precedence precedence) {
    advance_to_next_token(context);
    NudParseFunction prefix_rule = get_parse_rule(get_previous_token(context))->nud;

    if (prefix_rule == NULL) {
        error_at_previous_token(context, "Expect expression");
    }

    AstNode* node = prefix_rule(context);

    while (precedence < get_precedence(get_current_token(context))) {
        advance_to_next_token(context);
        LedParseFunction infix_rule = get_parse_rule(get_previous_token(context))->led;
        node = infix_rule(context, node);
    }

    return node;
}

static AstNode* parse_expression(ParsingContext* context) {
    return parse_precedence(context, PREC_ASSIGNMENT);
}

static AstNode* parse_number_literal_expression(ParsingContext* context) {
    Token* token = get_previous_token(context);

    if (token_is_double(token)) {
        return ast_node_make_double(token);
    } else {
        return ast_node_make_integer(token);
    }
}

static AstNode* parse_binary_operator_expression(ParsingContext* context, AstNode* left) {
    Token* previous = get_previous_token(context);
    ParseRule* rule = get_parse_rule(previous);
    AstNode* right = parse_precedence(context, rule->precedence);
    return ast_node_make_binary_infix_expression(previous, left, right);
}

static AstNode* parse_unary_prefix_operator_expression(ParsingContext* context) {
    Token* previous = get_previous_token(context);
    AstNode* right = parse_precedence(context, PREC_UNARY_PREFIX - 1);
    return ast_node_make_unary_prefix_expression(previous, right);
}

static AstNode* parse_grouping_expression(ParsingContext* context) {
    token_free(get_previous_token(context));
    AstNode* node = parse_expression(context);
    consume_current_token_if_type(context, TOKEN_RIGHT_PAREN, "Expect ')' after expression");
    return node;
}

static ParseRule parse_rules[] = {
        {parse_grouping_expression,       NULL,                                    PREC_CALL},    // TOKEN_LEFT_PAREN,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_RIGHT_PAREN,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_LEFT_BRACE,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_RIGHT_BRACE,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_COMMA,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_DOT,
        {parse_unary_prefix_operator_expression, parse_binary_operator_expression, PREC_TERM},    // TOKEN_MINUS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_MINUS_MINUS,
        {NULL,                                   parse_binary_operator_expression, PREC_TERM},    // TOKEN_PLUS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_PLUS_PLUS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_SEMICOLON,
        {NULL,                                   parse_binary_operator_expression, PREC_FACTOR},  // TOKEN_SLASH,
        {NULL,                                   parse_binary_operator_expression, PREC_FACTOR},  // TOKEN_STAR,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_QUESTION,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_COLON,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_AND_AND,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_PIPE_PIPE,

        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_BANG,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_BANG_EQUAL,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_EQUAL,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_EQUAL_EQUAL,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_GREATER,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_GREATER_EQUAL,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_LESS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_LESS_EQUAL,

        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_IDENTIFIER,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_STRING,
        {parse_number_literal_expression, NULL,                                    PREC_NONE},    // TOKEN_NUMBER,

        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_CLASS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_ELSE,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_FALSE,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_FUNC,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_FOR,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_IF,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_NIL,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_PRINT,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_RETURN,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_SUPER,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_THIS,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_TRUE,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_VAR,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_WHILE,

        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_ERROR,
        {NULL,                            NULL,                                    PREC_NONE},    // TOKEN_EOF
};

static ParseRule* get_parse_rule(Token* token) {
    return &parse_rules[token->type];
}

static Precedence get_precedence(Token* token) {
    ParseRule* rule = get_parse_rule(token);
    return rule->precedence;
}

AstNode* parser_parse(const char* source) {
    ParsingContext context = {
            .lexer = {

            },
            .parser = {

            }
    };

    lexer_init(&context.lexer, source);
    parser_init(&context.parser);

    advance_to_next_token(&context);

    AstNode* expression = parse_expression(&context);

    parser_free(&context.parser);
    lexer_free(&context.lexer);

    return expression;
}
